/**
 * @class
 */
define([
    'sna/model/ApplicationContextModel',
    'sna/model/widgets/LoadingSpinner/LoadingSpinner',
    'sna/networkview/controller/util/CollapseExpandedNodesHelper',
	'sna/networkview/controller/util/UndoStackHelper',
    'sna/util/GUIDUtil',
    'sna/networkview/controller/util/GroupsHelper',
    'sna/networkview/model/widgets/LayoutError/LayoutError',
    'sna/controller/ViewSwitcherController',
    './util/NetworkDataPreProcessor',
    'sna/util/PluginUtil',
    'sna/data/UserData',
    'sna/util/JSONUtil',
	"../../../commons/geom/Point",
    'sna/networkview/model/util/NLDHelper',
    './util/NetworkStatefulHelper',
    'dojo/query',
    'dojo/on',
    'graphics/gtl/core/nodeLinkDiagram/LinkDirectionEnum',
    'dojo/Stateful',
    'graphics/gtl/core/nodeLinkDiagram/NodeLinkDiagramItemTypeEnum',
    'sna/networkview/model/widgets/NetworkInfoDump/NetworkInfoDump/NetworkInfoDump',
    'sna/util/I18nUtil',
    'dijit/form/Button',
    'sna/util/DebugUtil',
    'dijit/registry',
	'dojo/dom-construct',
    'sna/util/MessageUtil',
    'sna/util/UserPreferenceUtil',
    './util/TimeSliderHelper',
    './util/NodeLinkFilterHelper',
    './util/CustomMeasuresHelper',
    'sna/util/ContextPreferenceUtil',
    'dojo/_base/lang',
    'sna/data/NetworkData',
    'dojo/_base/declare',
    'sna/networkview/model/NetworkModel',
    'dojox/lang/functional'
], function(ApplicationContextModel, LoadingSpinner, CollapseExpandedNodesHelper, UndoStackHelper, GUIDUtil, GroupsHelper, LayoutError, ViewSwitcherController,
	         NetworkDataPreProcessor, PluginUtil, UserData, JSONUtil, Point, NLDHelper, NetworkStatefulHelper, query, on,
	         LinkDirectionEnum, Stateful, NodeLinkDiagramItemTypeEnum, NetworkInfoDump, I18nUtil, Button, DebugUtil, registry,
             domConstruct, MessageUtil, UserPreferenceUtil, TimeSliderHelper, NodeLinkFilterHelper, CustomMeasuresHelper, 
             ContextPreferenceUtil, lang, networkData, declare, NetworkModel, functional ) {
    var userPreferenceKey = 'networkgraph';
	var map;
	var divmap;
	var dialog;
	
	// used when creating new network entities
	var PrototypeEntity = {
	    emptyNode: {
			"uid":"",
	        "positionX": (Math.random() * 200 + 300),
	        "positionY": (Math.random() * 200 + 300),
	        "nodeLabel":"",
	        "nodeColor":"",
	        "nodeType":"",
	        "nodeSymbol":"",
	        "nodeToolTip":"",
	        "isNodeHidden":false,
	        "nodeDegrees":0,
	        "borderSize":1,
	        "type":"NODE",
	        "evolves":false,
	        "evolutions2":{
	        },
	        "isDirty":false,
	        "customMeasures": {
	        }	
	    },
	    
	    emptyLink: {
	        clusterId: "",
	        fromNodeId: "",
	        toNodeId: "",
	        linkWidth: 2,
	        linkColor: "0x000000",
	        linkLabel: "",
	        linkStartDt: "", // May 1, 1998 12:00:00 AM
	        linkEndDt: "", // Oct 22, 2003 12:00:00 AM
	        type: "LINK",
	        strength: 20,
	        style: "A"
	    },

        emptyGroup: {
            uid: "",
            nodeUid: "",
            groupColor: "0x00cc00",
            groupSymbol: "User_2",
            isNodeHidden: false,
            borderSize: 2,
            type: "GROUP"
        }
	};
	
	var getEntityFromUID = function( uid ) {
		var entity = null;

		// see if it's a node
		functional.forIn(this.networkState.nodes, function( node, index ) {
			if( node.uid === uid ) {
				entity = node;
				return;
			}
		});
		
		// see if it's a link
		if( !entity ) {
			functional.forIn(this.networkState.links, function( link, index ) {
				if( link.uid === uid ) {
					entity = link;
					return;
				}
			});
		}

		// see if it's a group
		if( !entity ) {
			functional.forIn(this.networkState.groups, function( group, index ) {
				if( group.uid === uid ) {
					entity = group;
					return;
				}
			});
		}

		return entity;
	};
	
	// return the currently focused entity (node or link)
	var getCurrentlyFocusedEntity = function() {
        var currentlyFocusedEntity = undefined;
		
		var that = this;
		var currentlyFocusedUid = that.networkState.focusedEntityId; 
		
		// see if it's a node
		functional.forIn(this.networkState.nodes, function( node, index ) {
			if( node.uid === currentlyFocusedUid ) {
				currentlyFocusedEntity = node;
				return;
			}
		});
		
		// see if it's a link
		if( !currentlyFocusedEntity ) {
			functional.forIn(this.networkState.links, function( link, index ) {
				if( link.uid === currentlyFocusedUid ) {
					currentlyFocusedEntity = link;
					return;
				}
			});
		}
		
		// see if it's a group
		if( !currentlyFocusedEntity ) {
			functional.forIn(this.networkState.groups, function( group, index ) {
				if( group.uid === currentlyFocusedUid ) {
					currentlyFocusedEntity = group;
					return;
				}
			});
		}
		
		return currentlyFocusedEntity || {};
	};
	
	var updateTimeSlider = function() {
		// generate timeslider values, and pass them to the model to be applied to the timeslider component
        var timeSliderValues = TimeSliderHelper.getTimeSliderValues(this.networkState);
        this.model.updateTimeSlider(timeSliderValues);
	};
	
	// when isBisected is true, disable hierarchical layout option
    // update: we no longer disable the layout buttons so we don't have to maintain
    // the isbisected property when people add nodes/links/etc.
    // this is kept here in case any other action is required
	var isBisectedAction = function( isBisected ) {

	};

    var removeSingleGroups = function ( groups ) {
            var newGroups = new Array();
            for (var i = 0; i < groups.length; i++) {
                for (var j = 0; j < groups.length; j++) {
                    if (groups[i].uid === groups[j].uid && i != j) {
                        newGroups.push(groups[i]);
                        break;
                    }
                }
            }
            return newGroups;
        };
	
	// when isSaveNetworkAllowed is false, disable save button
	var isSaveNetworkAllowedAction = function( isSaveNetworkAllowed ) {
        var saveButton = registry.byId('snaSaveButton');
		if( isSaveNetworkAllowed === true ) {
			saveButton.set('disabled', false);
            ApplicationContextModel.set('unsavedChanges', true);
        }
		else {
			saveButton.set('disabled', true);
            ApplicationContextModel.set('unsavedChanges', false);
        }
	};
	
	var doesLinkAlreadyExist = function( node1, node2 ) {
		var alreadyExists = false;
		
		var nodeUid1 = node1.uid;
		var nodeUid2 = node2.uid;
		
		functional.forIn(this.networkState.links, function( link, linkIndex ) {
			var fromNodeId = link.fromNodeId;
			var toNodeId = link.toNodeId;
			
			var fromNodeIdMatches = (fromNodeId === nodeUid1 || fromNodeId === nodeUid2);
			var toNodeIdMatches = (toNodeId === nodeUid1 || toNodeId === nodeUid2);
			
			if( fromNodeIdMatches && toNodeIdMatches ) {
				alreadyExists = true;
			} });
		
		return alreadyExists;
	};

	var arrayMinusUid = function( array, field, uid ) {
		var subsetArray = [];
		
		functional.forIn(array, function( entity, index ) {
			if( entity[field] !== uid ) {
			    subsetArray.push(entity);
			}
		});
		
		return subsetArray;
	};

	return declare([], {
		constructor: function( args ) {
	        this.networkFacetController = args.networkFacetController;
	        this.networkPropertiesController = args.networkPropertiesController;
	        this.networkDetailsController = args.networkDetailsController;
            this.unchartedNodesController = args.unchartedNodesController;
            this.networkMapController = args.networkMapController;
            this.viewSwitcherController = args.viewSwitcherController;
            this.mode = this.MODES.FULL_SNA;

            networkData.linkPreProcessor = NetworkDataPreProcessor.preprocessLinks;
            networkData.nodePreProcessor = NetworkDataPreProcessor.preprocessNodes;
	    },

        MODES: {
            'FULL_SNA': 0,
            'SNALITE': 1
        },

        setMode: function(mode) {
            // whether we are in sna lite or full blown sna
            if(mode === 'SNALITE') {
                this.mode = this.MODES.SNALITE;
            }
            else {
                this.mode = this.MODES.FULL_SNA;
            }
        },
	    
	    updateEntity: function( uid, propertyMap ) {
	    	var that = this;
	    	var getJsonEntity = function( entityUid ) {
	    		var entity = null;
	    		
	    		// check nodes
	    		var nodes = that.networkState.nodes;
	    		for( var i = 0; i < nodes.length; i++ ) {
	    			if( nodes[i] && nodes[i].uid === entityUid ) {
	    				entity = nodes[i];
	    				break;
	    			}
	    		}
	    		
	    		if( !entity ) {
	    			// check links
	    			var links = that.networkState.links;
	    			for( var i = 0; i < links.length; i++ ) {
	    				if( links[i] && links[i].uid === entityUid ) {
	    					entity = links[i];
	    					break;
	    				}
	    			}
	    		}

	    		if( !entity ) {
	    			// check groups
	    			var groups = that.networkState.groups;
	    			for( var i = 0; i < groups.length; i++ ) {
	    				if( groups[i] && groups[i].uid === entityUid ) {
	    					entity = groups[i];
	    					break;
	    				}
	    			}
	    		}
	    		
	    		return entity || {};
	    	};
	    	
	    	var jsonEntity = getJsonEntity(uid);
	    	
	    	lang.mixin(jsonEntity, propertyMap);
	    	
	    	NetworkStatefulHelper.sync([jsonEntity]);
	    },

	    /**
	     * Start main network execution
	     * 
	     * vertical: the analytic context
	     * entity: the actionable entity
	        fields:
	          actionableEntityDescription
			  actionableEntityId
			  actionableEntityName
			  actionableEntityRank
			  actionableEntityType
			  alertDescription 
			  alertId
			  alertRank
			  alertScenario
			  alertScore
			  alertStatus
			  alertTriggerDttm
			  alertType
			  entityId
			  entityName
			  entityType
			  entityValidFromDttm
			  entityValidToDttm
			  generatedCommentId
			  
	     * d_parentId: the dom node to attach view to
	     */
	    start: function( vertical, entity, d_parentId ) {
	    	UndoStackHelper.clearStack(UndoStackHelper.GROW);
	    	UndoStackHelper.clearStack(UndoStackHelper.REGROUP);

            dialog = new LayoutError({});

	    	// NetworkModel manages the nodelink component
	    	this.model = new NetworkModel({
            	controller: this
            });
	    	
	    	if( this.model.isAlreadyLoaded() === true ) {
	    		return;
	    	}
	    	
	    	if( !d_parentId ) {
	    		console.warn('NetworkController.start: d_parentId is undefined');
	    	}
	    	this.d_parentId = d_parentId;

	    	// holds the network state
	    	this.networkState = {
	    		vertical: vertical,
	    		entity: entity,
	    		nodes: {},
	    		links: {},
	    		facets: {},
	    		groups: [],
	    		unchartedNodes: [],
	    		focusedEntityId: -1,
	    		actionableEntityId: entity.actionableEntityId,
	    		actionableEntityName: entity.actionableEntityName,
	    		actionableEntityType: entity.actionableEntityType
	    	};
	    	
	    	// after the network view dom is loaded, get the network data and start setting up view and state
	    	var that = this;
	    	this.model.load(function() {
	    		// start network controllers
		    	that.networkPropertiesController.start();
		    	that.networkDetailsController.start();
                that.networkMapController.start();

                var complete = false;
                var completionCallback = function() {
                    return complete;
                };
                new LoadingSpinner().target(registry.byId('snadiagramcontentpane')).until(completionCallback).startSpinners(function(){});

                // get network data and network plugins
		    	var preferredSource = 'NETWORK_SOURCE_DATABASE';
                // TODO this is duplicated by reloadNetwork
		    	networkData.getSocialNetworkAndFacets(vertical, that.networkState.entity.actionableEntityId, that.networkState.entity.actionableEntityType, preferredSource, function( results ) {
		    		UserData.getUser( function( userData ) {
                        complete = true;

                        lang.mixin(that.networkState, results);

                        isBisectedAction(results.isBisected);
                        isSaveNetworkAllowedAction(results.isSaveNetworkAllowed);

                        var justNodesAndLinks = results.nodes;
                        justNodesAndLinks = justNodesAndLinks.concat(results.links);
                        if (results.groups != null) {
                            justNodesAndLinks = justNodesAndLinks.concat(removeSingleGroups(results.groups));
                        }
                        NetworkStatefulHelper.clear();
                        NetworkStatefulHelper.sync(justNodesAndLinks);

                        that.model.render(NetworkStatefulHelper.getStatefulEntityArray(), results.plugins, vertical, d_parentId, function() {
                            // generate timeslider values, and pass them to the model to be applied to the timeslider component
                            var timeSliderValues = TimeSliderHelper.getTimeSliderValues(that.networkState);
                            that.model.setupTimeSlider(timeSliderValues);

                            // load hidden plugins after the NLD has loaded
                            PluginUtil.createPluginsFromMap(results.plugins, {}, vertical);

                            NetworkStatefulHelper.handleRawGroups(results);

                            // now we can generate the facets since we have the node's custom measures data
                            that.networkFacetController.start(that.networkState.facets);
	                        that.networkFacetController.refresh(that.networkState.facets); 
                        });
		    		});
	            });
		    	
		    	if( DebugUtil.isShowNetworkInfoMode() ) {
			    	var networktoolbar = registry.byId('networktoolbar');
			    	networktoolbar.addChild(new Button({
			    		id: 'snaShowNetworkInfoButton',
			    		iconClass: 'sasPropertiesIcon', 
			    		showLabel: false, 
			    		label: I18nUtil.get('networkview_shownetworkinformation')
			    	}));
		    	}
	    	});
	    },
	    
	    /**
	     * 
	     */
	    startAddNodeActions: function() {
	    	var that = this;
	    	
	    	var currentlyFocusedEntity = lang.hitch(this, getCurrentlyFocusedEntity)();
	    	if( currentlyFocusedEntity.type === 'NODE' ) {

                var newPrototypeNode = {};
                lang.mixin(newPrototypeNode, PrototypeEntity.emptyNode);

                var newPrototypeLink = {};
                lang.mixin(newPrototypeLink, PrototypeEntity.emptyLink);

		    	this.model.showAddNodeDialog(this.networkState, currentlyFocusedEntity, newPrototypeNode, newPrototypeLink, function( newNode, newLink ) {
		    		that.networkState.nodes.push(newNode);
		    		that.networkState.links.push(newLink);
		    		
		    		NetworkStatefulHelper.sync([newNode, newLink]);
		    		
		    		MessageUtil.showConfirmMessage(I18nUtil.get('networkview_nodeadded'));
		    		lang.hitch(that, updateTimeSlider)();
		    		
		    		//this refreshes the timeslider so that the new start and ends dates get taken in account for
		    		that.filterDataFromTimeSliderValue(that.model.getTimeSliderValue());

		    		// handle collapse expanded nodes undo stack
	    			UndoStackHelper.clearStackIfContains(UndoStackHelper.GROW, [
	    				currentlyFocusedEntity.uid
	    			]);

                    // update map icons if a node with lat/long coords was added
                    that.networkMapController.updateMenuState();
		    	});
	    	}
	    	else {
	    		MessageUtil.showErrorMessage(I18nUtil.get('networkview_youneedtohaveanodeselected'));
	    	}
	    },
	    
	    /**
	     * 
	     */
	    startEditEntityActions: function() {
	    	var that = this;
	    	
	    	var currentlySelectedIndicies = NLDHelper.getCurrentlySelectedIndicies();
	    	var currentlySelectedEntities = NetworkStatefulHelper.entitiesFromIndicies(currentlySelectedIndicies);
	    	
	    	if( currentlySelectedEntities.length === 0 ) {
	    		MessageUtil.showErrorMessage(I18nUtil.get('networkview_selectanentity'));
	    	}
	    	else if( currentlySelectedEntities.length > 1 ) {
	    		MessageUtil.showErrorMessage(I18nUtil.get('networkview_pleaseselectonlyoneentitytoedit'));
	    	}
	    	else if( currentlySelectedEntities.length === 1 ) {
	    		var currentlyFocusedEntity = currentlySelectedEntities[0];
	    		var isNode = currentlyFocusedEntity.type._id === 'NODE';
	    		var isLink = currentlyFocusedEntity.type._id === 'LINK';

                var jsonEntity = lang.hitch(this, getEntityFromUID, currentlyFocusedEntity.get('uid'))();
	    		
	    		if( isNode === true ) {
			    	this.model.showEditNodeDialog(jsonEntity, function( updatedNode ) {
			            // update node positions before submitting
			            NetworkStatefulHelper.reverseSync(that.networkState.nodes);

			    		NodeLinkFilterHelper.updateNode(jsonEntity, updatedNode);
			    		
			    		MessageUtil.showConfirmMessage(I18nUtil.get('networkview_nodeupdated'));
			    		lang.hitch(that, updateTimeSlider)();
			    		that.updateEntitiesFromTimesliderValue();
			    	});
		    	}
		    	else if( isLink === true ) {
			    	this.model.showEditLinkDialog(this.networkState, jsonEntity, function( updatedLink ) {
			            // update node positions before submitting
			            NetworkStatefulHelper.reverseSync(that.networkState.nodes);
			            
			    		MessageUtil.showConfirmMessage(I18nUtil.get('networkview_linkupdated'));
                        NetworkStatefulHelper.sync([updatedLink]);
			    		lang.hitch(that, updateTimeSlider)();
			    		that.updateEntitiesFromTimesliderValue();
			    	});
		    	}
	    	}
	    },
	    
	    /**
	     * 
	     */
	    startAddLinkActions: function() {
	    	var that = this;
	    	
	    	var currentlySelectedIndicies = NLDHelper.getCurrentlySelectedIndicies();
	    	var currentlySelectedEntities = NetworkStatefulHelper.entitiesFromIndicies(currentlySelectedIndicies);
	    	
	    	var selectedNodes = [];
	    	functional.forIn(currentlySelectedEntities, function( entity, entityIndex ) {
	    		if( entity.type._id === 'NODE' ) {
	    			selectedNodes.push(entity);
	    		}
	    	});
	    	
	    	if( selectedNodes.length === 2 ) {
	    		var node0 = selectedNodes[0];
	    		var node1 = selectedNodes[1];
	    		if( lang.hitch(this, doesLinkAlreadyExist, node0, node1)() ) {
	    			MessageUtil.showErrorMessage(I18nUtil.get('networkview_alinkalreadyexistsbetweenthosetwonodes'));
	    		}
	    		else {
                    var newPrototypeLink = {};
                    lang.mixin(newPrototypeLink, PrototypeEntity.emptyLink);

		    		// start add link dialog
			    	this.model.showAddLinkDialog(this.networkState, node0, node1, newPrototypeLink, function( newLink ) {
			    		that.networkState.links.push(newLink);
			    		
			    		NetworkStatefulHelper.sync([newLink]);
			    		
			    		MessageUtil.showConfirmMessage(I18nUtil.get('networkview_linkadded'));
			    		lang.hitch(that, updateTimeSlider)();

			    		// handle collapse expanded nodes undo stack
		    			UndoStackHelper.clearStackIfContains(UndoStackHelper.GROW, [
		    				node0.uid,
		    				node1.uid
		    			]);
			    	});
	    		}
	    	}
	    	else if( selectedNodes.length > 2 ) {
	    		MessageUtil.showErrorMessage(I18nUtil.get('networkview_pleaseselectonlytwonodes'));
	    	}
	    	else {
	    		MessageUtil.showErrorMessage(I18nUtil.get('networkview_youneedtohavetwonodesselected'));
	    	}
	    },
	    
	    /**
	     * 
	     */
	    startRemoveEntityActions: function() {
	    	var that = this;
    	
	    	var currentlySelectedIndicies = NLDHelper.getCurrentlySelectedIndicies();
	    	var currentlySelectedEntities = NetworkStatefulHelper.entitiesFromIndicies(currentlySelectedIndicies);
	    	
	    	if( this.networkState.nodes.length === 1 ) {
	    		MessageUtil.showErrorMessage(I18nUtil.get('networkview_networkmusthaveatleastonenode'));
	    	}
	    	else if( currentlySelectedEntities.length === 0 ) {
	    		MessageUtil.showErrorMessage(I18nUtil.get('networkview_selectanentity'));
	    	}
	    	else if( currentlySelectedEntities.length > 1 ) {
	    		MessageUtil.showErrorMessage(I18nUtil.get('networkview_pleaseselectonlyoneentitytoremove'));
	    	}
	    	else if( currentlySelectedEntities.length === 1 ) {
	    		var currentlyFocusedEntity = currentlySelectedEntities[0];

	    		var isNode = currentlyFocusedEntity.type._id === 'NODE';
	    		var isLink = currentlyFocusedEntity.type._id === 'LINK';
	    		this.willDeletionBisectGraph(currentlyFocusedEntity.get('uid'), function( willBisect ) {
					if( willBisect === true ) {
						MessageUtil.showErrorMessage(I18nUtil.get('networkview_cannotperformdeletionwillcreatebisectedgraph'));
					}
					else {
						if( isNode === true ) {
			    			NodeLinkFilterHelper.removeNode(that.networkState, currentlyFocusedEntity.uid);
			    			MessageUtil.showConfirmMessage(I18nUtil.get('networkview_noderemoved'));

			    			UndoStackHelper.clearStackIfContains(UndoStackHelper.GROW, [currentlyFocusedEntity.uid]);
			    			UndoStackHelper.clearStackIfContains(UndoStackHelper.REGROUP, [currentlyFocusedEntity.uid]);
			    		}
			    		else if( isLink === true ) {
			    			NodeLinkFilterHelper.removeLink(that.networkState, currentlyFocusedEntity.uid);
			    			MessageUtil.showConfirmMessage(I18nUtil.get('networkview_linkremoved'));

			    			UndoStackHelper.clearStackIfContains(UndoStackHelper.GROW, [
			    				currentlyFocusedEntity.toNodeId,
			    				currentlyFocusedEntity.fromNodeId
			    			]);
			    			UndoStackHelper.clearStackIfContains(UndoStackHelper.REGROUP, [
			    				currentlyFocusedEntity.toNodeId,
			    				currentlyFocusedEntity.fromNodeId
			    			]);
			    		}
			    	
						// hide properties and details pane, since the currently selected entity is removed
                        that.hideNodeSpecificSections();
					}
	    		});
	    	}
	    },

        addGroup: function( groupUid, nodeUidArray ) {
            var that = this;

            var arrayContainsGroupUid = function() {
            	var containsGroupUid = false;
            	for( var i = 0; i < nodeUidArray.length; i++ ) {
            		var uid = nodeUidArray[i];

            		var entity = lang.hitch(that, getEntityFromUID, uid)();

            		if( entity.type === 'GROUP' ) {
            			containsGroupUid = true;
            			break;
            		}
            	}

            	return containsGroupUid;
            };

            if( nodeUidArray.length < 2 ) {
            	MessageUtil.showErrorMessage(I18nUtil.get('networkview_theremustbetwoormorenodesselected'));
            }
            else if( arrayContainsGroupUid() === true ) {
            	MessageUtil.showErrorMessage(I18nUtil.get('networkview_groupscannotbenested'));
            }
            else {
	            var jsonGroupArray = [];
	            if( !groupUid ) {
	            	groupUid = GUIDUtil.generate();
	        	}
	            functional.forIn(nodeUidArray, function( nodeUid, nodeUidIndex ) {
	            	var proposedEntity = lang.hitch(that, getEntityFromUID, nodeUid)();
	            	if( proposedEntity && proposedEntity.type === 'NODE' ) {
		                var newJsonGroup = {};
		                lang.mixin(newJsonGroup, PrototypeEntity.emptyGroup);
	
		                newJsonGroup.uid = groupUid;
		                newJsonGroup.nodeUid = nodeUid;
	
		                that.networkState.groups.push(newJsonGroup);
	
		                jsonGroupArray.push(newJsonGroup);
	            	}
	            });

	            NetworkStatefulHelper.addGroup(groupUid, jsonGroupArray);
	        }
        },

        removeGroup: function( groupUid ) {
            var that = this;
            functional.forIn(this.networkState.groups, function( group, groupIndex ) {
                if( group.uid === groupUid ) {
                    delete that.networkState.groups[groupIndex];
                }
            });

            NetworkStatefulHelper.removeGroup(groupUid);
            this.hideNodeSpecificSections();
        },

        hideNodeSpecificSections: function() {
            this.networkPropertiesController.hide();
            this.networkDetailsController.hide();
        },
	    
	    /**
	     * Cleanup view 
	     * used when switching to a different view
	     */
	    cleanup: function() {
            if( model ) {
                model.cleanupView();
            }
        },
        
        /**
         * 
         */
        setFocusedEntityId: function( entityId ) {
        	this.networkState.focusedEntityId = entityId;
        },

        /**
         *
         */
        getFocusedEntityId: function( entityId ) {
            return this.networkState.focusedEntityId;
        },

        /**
         * 
         */
	    showPropertiesSection: function(previousEntityId) {
            var that = this;
	    	var currentlyFocusedEntity = lang.hitch(this, getCurrentlyFocusedEntity)();

	    	if( currentlyFocusedEntity ) {
	    		// open properties section for focused entity
                if (previousEntityId !== this.networkState.focusedEntityId){
		    	    this.networkPropertiesController.showEntityProperties(currentlyFocusedEntity, this.networkState);
                }

		    	// hide details section if focused entity is a group
		    	if( currentlyFocusedEntity.type === 'GROUP' ) {
		    		this.networkDetailsController.hide();
		    	}
		    	// else, if the details section is already open, update it
		    	else if( this.networkDetailsController.isOpen() && currentlyFocusedEntity.type === 'NODE'
                    && (previousEntityId !== this.networkState.focusedEntityId)) {
                    // this is called by enough people that it warrants being handled here instead of individual locations
                    var loadingComplete = false;
                    var completionFunction = function() {
                        return loadingComplete;
                    };
                    this.networkPropertiesController.startLoadingSpinner(completionFunction);
                    this.networkDetailsController.startLoadingSpinner(completionFunction);

		    		this.showDetails(currentlyFocusedEntity, function() {
                        loadingComplete = true;
                    }, function() {
                        loadingComplete = true; //error
                        that.networkDetailsController.showLoadingError();
                    });
		    	}
                else if (currentlyFocusedEntity.type === 'LINK') {
                    this.networkDetailsController.hide();
                }
		    	else {
		    		// TODO: resize NLD component?
		    	}
	    	}
	    },
	    
	    /**
	     * 
	     */
	    isPropertiesSectionOpen: function() {
	    	return this.networkPropertiesController.isOpen();
	    },
		    
		/**
		 * 
		 */
	    startHighlightNetworkActions: function( facets ) {
	    	var that = this;
		    CustomMeasuresHelper.performAspectTracing(this.networkState, facets, function( entity, highlight ) {
		    	that.aspectTracingAction(entity, highlight);
		    });
	    },
	    
	    /**
	     * tell networkDetailsController to show details for the node with the given id
	     */
        showDetails: function(entity, callback, errorCallback) {
	    	var that = this;
        	this.networkDetailsController.show(this.networkState.focusedEntityId, entity, function() {
        		// TODO: resize NLD component?
                if(callback) {
                    callback();
                }
        	}, errorCallback);
        },
        
        /**
         * 
         */
        filterDataFromTimeSliderValue: function( timeSliderValue ) {
        	var timesliderAlgorithm = this.model.getTimeSliderAlgorithm();
        	
        	var that = this;
        	TimeSliderHelper.filterDataBasedOnTimeSliderValue(this.networkState, timeSliderValue, timesliderAlgorithm, {
        		nodeChangeDateAction: function() {
        			that.updateEntitiesFromTimesliderValue();
        		},
        		linkDimAction: function( link, dim ) {
        			that.timeSliderLinkDimAction(link, dim);
        		},
        		nodeDimAction: function( node, dim ) {
        			that.timeSliderNodeDimAction(node, dim);
        		}
        	});
        }, 
        
        /**
         * 
         */
        saveNetwork: function() {
        	var fieldsToSave = [
                'uid', 
                'positionX', 
                'positionY', 
                'nodeLabel',
                'nodeType',
                'evolutions2',
                'nodeToolTip',
                "isNodeHidden",
                "nodeDegrees",
                "borderSize",
                "type",
                "evolves",
                'latitude',
                'longitude',
                "linkDirection"
            ];
        	
        	NetworkStatefulHelper.reverseSync(this.networkState.nodes);
        	
        	var subsetNodes = JSONUtil.getSubsetObjectArray(this.networkState.nodes, fieldsToSave);
        	
        	var customMeasures = {};
        	if( this.networkState.tables ) {
        		customMeasures = this.networkState.tables.MEASURES;
        	}
        	
        	var modifiedLinks = lang.clone(this.networkState.links); 
        	functional.forIn(modifiedLinks, function(link) {
                // since vector is deprecated, still have to persist it
                if(link.linkDirection && link.linkDirection === 'TO') {
                    link.type = 'VECTOR';
                }
                else {
                    link.type = 'LINK';
                }

        		if(link.linkStartDt) {
        			if(link.linkStartDt.getTime) {
        				link.linkStartDt = link.linkStartDt.getTime();	
        			}
        			else {
        				link.linkStartDt = new Date(link.linkStartDt).getTime(); 
        			}
        		}
        		
        		if(link.linkEndDt) {
        			if(link.linkEndDt.getTime) {
        				link.linkEndDt = link.linkEndDt.getTime();	
        			}
        			else {
        				link.linkEndDt = new Date(link.linkEndDt).getTime(); 
        			}
        		}
        	});
        	
        	networkData.saveNetwork(
        		this.networkState.actionableEntityId, 
        		this.networkState.actionableEntityType,
        		subsetNodes,
        		modifiedLinks,
        		this.networkState.groups,
        		customMeasures, 
        		function( data ) {
        			if( data.success === 'true' || data.success === true ) {
        				MessageUtil.showConfirmMessage(I18nUtil.get('networkview_networksaved'));
        			}
        			else {
        				MessageUtil.showErrorMessage(I18nUtil.get('networkview_saveunsuccessfulpleaseseeadministrator'));
        			}
        		}
        	);
        },
        
        /**
         * 
         */
        showNetworkInfoButtonClicked: function() {
        	var nid = new NetworkInfoDump();
        	nid.print(this.networkState, NetworkStatefulHelper.getStatefulEntityArray());
        },

        /**
         * 
         */
        attachAnnotation: function( toolTip ) {
			var currentlyFocusedEntity = lang.hitch(this, getCurrentlyFocusedEntity)();
			
			if( currentlyFocusedEntity.type === 'NODE' ) {
				currentlyFocusedEntity.nodeToolTip = toolTip;
				MessageUtil.showConfirmMessage(I18nUtil.get('networkview_annotationattachedtonode'));
			}
            else if( currentlyFocusedEntity.type === 'LINK' ) {
                currentlyFocusedEntity.linkLabel = toolTip;
                MessageUtil.showConfirmMessage(I18nUtil.get('networkview_annotationattachedtolink'));
            }
		},

        /**
         *
         */
        attachAnnotationToNode: function( uid, toolTip ) {
            var entity = lang.hitch(this, getEntityFromUID, uid)();

            if( entity.type === 'NODE' ) {
                entity.nodeToolTip = toolTip;
                MessageUtil.showConfirmMessage(I18nUtil.get('networkview_annotationattachedtonode'));
            }
        },

        getAnnotation: function( uid ) {
            var annotation = null;

            var entity = lang.hitch(this, getEntityFromUID, uid)();
            if( entity ) {
                if( entity.type === 'NODE' ) {
                    annotation = entity.nodeToolTip;
                }
                else if( entity.type === 'LINK' ) {
                    annotation = entity.linkLabel;
                }
            }

            return annotation;
        },
		
		/**
		 * 
		 */
		timeSliderNodeChangeDateAction: function( node, evolutionsIndex ) {
			var getLatestChangeDate = function( node, changeDate ) {
				var latest = "0";
				
				functional.forIn(node.evolutions2, function( evolution, evolutionIndex ) {
					var currentEvolution = Number(evolutionIndex);
					var nChangeDate = Number(changeDate);
					
					if( currentEvolution <= nChangeDate) {
						latest = ('' + currentEvolution);
					}
				});
				
				return latest;
			};
			
			var latestChangeDate = getLatestChangeDate(node, evolutionsIndex);
			NodeLinkFilterHelper.updateNode(node, node.evolutions2[latestChangeDate]);
			this.updateEntity(node.uid, node);
		},

        /**
         * perform actions on network entities such as dimming
         * with regard to the current timeslider value
         */
		updateEntitiesFromTimesliderValue: function() {
			var that = this;
			var timesliderValue = this.model.getTimeSliderValue();
			functional.forIn(this.networkState.nodes, function( node, nodeIndex ) {
				that.timeSliderNodeChangeDateAction(node, timesliderValue);
			});
		},
		
		/**
		 * 
		 */
		timeSliderLinkDimAction: function( link, dim ) {
			var statefulArrayIndex = NetworkStatefulHelper.getArrayIndex(link.uid);
            // TODO why is this dim and not !dim like below???
			NLDHelper.dimEntity(statefulArrayIndex, dim);
		},
		
		/**
		 * 
		 */
		timeSliderNodeDimAction: function( node, dim ) {
			var statefulArrayIndex = NetworkStatefulHelper.getArrayIndex(node.uid);
			NLDHelper.dimEntity(statefulArrayIndex, !dim);
		},

		/**
		 * 
		 */
		aspectTracingAction: function( entity, highlight ) {
			this.updateEntity(entity.uid, {
				highlight: highlight
			});
		},
		
		/**
		 * 
		 */
		startRunLayoutAlgorithmActions: function( algorithmString, params, completionCallback ) {
			var that = this;
			var subsetNodes = JSONUtil.getSubsetObjectArray(this.networkState.nodes, ['uid']);
			var subsetLinks = JSONUtil.getSubsetObjectArray(this.networkState.links, ['uid', 'fromNodeId', 'toNodeId']);
			networkData.getLayout(subsetNodes, subsetLinks, algorithmString, params, function( nodes ) {
				functional.forIn(nodes, function( node, nodeIndex ) {
					that.updateEntity(node.uid, {
		        		positionX: node.positionX,
		        		positionY: node.positionY
		        	});
				});
				
				NLDHelper.center();

				if (completionCallback) {
					completionCallback();
					
					if (nodes === "") { //if graph is null
			            dialog.showDialog();
					}
				}
			});
		},
		
		/**
	     * 
	     */
	    startGrowNetworkActions: function( nodeUid, nodeType ) {
			var that = this;

            // update node positions before submitting
            NetworkStatefulHelper.reverseSync(this.networkState.nodes);

	    	var subsetNodes = JSONUtil.getSubsetObjectArray(this.networkState.nodes, ['uid', 'positionX', 'positionY']);
			var subsetLinks = JSONUtil.getSubsetObjectArray(this.networkState.links, ['uid', 'fromNodeId', 'toNodeId']);

			networkData.growSocialNetwork(this.networkState.actionableEntityId, this.networkState.actionableEntityType, nodeUid, nodeType,  subsetNodes, subsetLinks, function( data ) {
				var isBisected = data.isBisected;
				isBisectedAction(isBisected);

				// shallow copy
            	var originalNodesAndLinks = that.networkState.nodes.slice(0);
            	originalNodesAndLinks = originalNodesAndLinks.concat(that.networkState.links.slice(0));

                // custom measures were regenerated
                that.networkState.tables = data.tables;

				// add/merge nodes
				functional.forIn(data.nodes, function( node, nodeIndex ) {
					var existingNode = lang.hitch(that, getEntityFromUID, node.uid)();

                    // if evolutions2 is {}, then this is just a position update for the node
					if( existingNode && !Object.keys(node.evolutions2).length === 0) {
						lang.mixin(existingNode, node);
					}
					else if (!existingNode) {
						that.networkState.nodes.push(node);
					}

                    // update the layout since it was ran
                    that.updateEntity(node.uid, {
                        positionX: node.positionX,
                        positionY: node.positionY
                    });
				});

				// add/merge links
				functional.forIn(data.links, function( link, linkIndex ) {
					var existingLink = lang.hitch(that, getEntityFromUID, link.uid)();
					
					if( existingLink ) {
						lang.mixin(existingLink, link);
					}
					else {
						that.networkState.links.push(link);
					}
				});

				var justNodesAndLinks = data.nodes;
                if( justNodesAndLinks ) {
                    justNodesAndLinks = justNodesAndLinks.concat(data.links);

                	var getNewNodesAndLinks = function( originalEntityArray, newEntityArray ) {
                		var differenceEntityArray = [];

                		var hasEntity = function( array, entity ) {
                			var hasE = false;

                			for( var j = 0; j < array.length; j++ ) {
                				if( array[j].uid === entity.uid ) {
                					hasE = true;
                					break;
                				}
                			}

                			return hasE;
                		};
                		for( var i = 0; i < newEntityArray.length; i++ ) {
                			var potentialNewEntity = newEntityArray[i];
                			if( hasEntity(originalEntityArray, potentialNewEntity) === false ) {
                				differenceEntityArray.push(potentialNewEntity);
                			}

                		}

                		return differenceEntityArray;
                	};

                    var newNodesAndLinks = getNewNodesAndLinks(originalNodesAndLinks, justNodesAndLinks);

                    NetworkStatefulHelper.sync(justNodesAndLinks);

                    // update facets
                    that.networkState.facets = data.facets;
                    that.networkFacetController.refresh(that.networkState.facets);

                    NLDHelper.center();

                    // generate timeslider values, and pass them to the model to be applied to the timeslider component
                    var timeSliderValues = TimeSliderHelper.getTimeSliderValues(that.networkState);
                    that.model.setupTimeSlider(timeSliderValues);
                    that.updateEntitiesFromTimesliderValue();

                    var newNodesAndLinksUids = [];
                    functional.forIn(newNodesAndLinks, function( newEntity ) {
                    	newNodesAndLinksUids.push(newEntity.uid);
                    });

					registry.byId('snaCollapseExpandedNodesButton').set('disabled', false);
					UndoStackHelper.push(UndoStackHelper.GROW, {
						keyUid: nodeUid,
						uids: newNodesAndLinksUids
					});

					var sourceNode = lang.hitch(that, getEntityFromUID, nodeUid)();
					sourceNode['degree'] = 0;
					NetworkStatefulHelper.sync([sourceNode]);

                    // update map button state if necessary
                    that.networkMapController.updateMenuState();
                }
                else {
                    console.warn("data.nodes is undefined")
                }
			});
	    },

	    collapseExpandedNodes: function() {
	    	CollapseExpandedNodesHelper.perform(this);
	    	//refresh timeslider
	    	this.filterDataFromTimeSliderValue(this.model.getTimeSliderValue());
	    },

        getGlobalSearchQueryFunctions: function() {
            return this.networkFacetController.getGlobalSearchQueryFunctions();
        },

        refreshNetworkDetailsGrids: function() {
            this.networkDetailsController.refreshDataGrids();
        },

	    /**
	     * Reload network
	     * source: 'NETWORK_SOURCE_STOREDPROCESS' or 'NETWORK_SOURCE_DATABASE'
	     */
	    reloadNetwork: function( source ) {
	    	// destroy networkdiagram div
	    	query(".networkdiagram").forEach(domConstruct.destroy);
	    	
	    	var that = this;
	    	
	    	// clear state
	    	this.networkState.nodes = {};
	    	this.networkState.links = {};
	    	this.networkState.groups = {};
	    	this.networkState.tables = {};
	    	this.networkState.facets = {};
	    	NetworkStatefulHelper.clear();

	    	UndoStackHelper.clearStack(UndoStackHelper.GROW);
	    	UndoStackHelper.clearStack(UndoStackHelper.REGROUP);
	    	
	    	var vertical = that.networkState.vertical;

            var complete = false;
            var completionCallback = function() {
                return complete;
            };
            new LoadingSpinner().target(registry.byId('snadiagramcontentpane')).until(completionCallback).startSpinners(function(){});

            // get network data and network plugins
	    	networkData.getSocialNetworkAndFacets(vertical, that.networkState.entity.actionableEntityId, that.networkState.entity.actionableEntityType, source, function( results ) {
	    		UserData.getUser( function( userData ) {
                    complete = true;

                    // hide properties and details pane, since the currently selected entity may not be the same
                    that.networkPropertiesController.hide();
                    that.networkDetailsController.hide();

                    lang.mixin(that.networkState, results);

                    isBisectedAction(results.isBisected);
                    isSaveNetworkAllowedAction(results.isSaveNetworkAllowed);

                    var justNodesAndLinks = results.nodes;
                    justNodesAndLinks = justNodesAndLinks.concat(results.links);
                    if (results.groups != null) {
                        justNodesAndLinks = justNodesAndLinks.concat(removeSingleGroups(results.groups));
                    }
                    NetworkStatefulHelper.sync(justNodesAndLinks);

                    that.model.render(NetworkStatefulHelper.getStatefulEntityArray(), results.plugins, vertical, that.d_parentId, function() {
                        // generate timeslider values, and pass them to the model to be applied to the timeslider component
                        var timeSliderValues = TimeSliderHelper.getTimeSliderValues(that.networkState);
                        that.model.setupTimeSlider(timeSliderValues);

                        // load hidden plugins after the NLD has loaded
                        PluginUtil.createPluginsFromMap(results.plugins, {}, vertical);
                        
                        that.updateEntitiesFromTimesliderValue();

                        NetworkStatefulHelper.handleRawGroups(results);

                        // now we can generate the facets since we have the node's custom measures data
                        that.networkFacetController.destroyFacets();
                        that.networkFacetController.start(that.networkState.facets);
                        that.networkFacetController.refresh(that.networkState.facets); 
                        
                        // disable group and collapse expanded nodes buttons after reload
                        registry.byId('snaRegroupNodesButton').set('disabled', 'disabled');
		                registry.byId('snaCollapseExpandedNodesButton').set('disabled', 'disabled');

                        // update the map menu items
                        that.networkMapController.updateMenuState();
                    });
	    		});
            });
	    },

        getEntityFromUID: function( uid ) {
            var entity = null;
            if( uid ) {
                entity = lang.hitch(this, getEntityFromUID, uid)();
            }

            return entity;
        },

	    willDeletionBisectGraph: function( uidToDelete, callback ) {
	    	var subsetNodes = JSONUtil.getSubsetObjectArray(this.networkState.nodes, ['uid']);
			var subsetLinks = JSONUtil.getSubsetObjectArray(this.networkState.links, ['fromNodeId', 'toNodeId']);
			
			var subsetNodesMinusEntity = arrayMinusUid(subsetNodes, 'uid', uidToDelete);
			var subsetLinksMinusFromNodeIdEntity = arrayMinusUid(subsetLinks, 'fromNodeId', uidToDelete);
	        var subsetLinksMinusEntity = arrayMinusUid(subsetLinksMinusFromNodeIdEntity, 'toNodeId', uidToDelete);

			networkData.deleteFromNetwork(subsetNodesMinusEntity, subsetLinksMinusEntity, function( data ) {
				var isBisected = data.isBisected;

				if( isBisected === true || isBisected === "true" ) {
					callback(true);
				}
				else {
					callback(false);
				}
			});
	    }
    });
});
